/*
 * Handler routines.
 *
 * Copyright (C) 2008 by Maurus Cuelenaere
 * Copyright (C) 2015 Paul Boddie <paul@boddie.org.uk>
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

.text
.extern irq_handle
.extern irq_counter
.globl _exc_stack
.globl real_exception_handler
.set noreorder
.set noat

#define C0_STATUS   $12
#define C0_CAUSE    $13
#define C0_EPC      $14
#define C0_CONFIG   $16
#define C0_ERROREPC $30
#define S_CauseExcCode  2
#define M_CauseExcCode  (0x1f << S_CauseExcCode)

_exc_stack:
    .word 0

real_exception_handler:
    addiu   $sp, -0x88      /* multiple of 8 */
    sw      $ra,    0($sp)
    sw      $fp,    4($sp)
    sw      $gp,    8($sp)
    sw      $t9,  0xC($sp)
    sw      $t8, 0x10($sp)
    sw      $s7, 0x14($sp)
    sw      $s6, 0x18($sp)
    sw      $s5, 0x1C($sp)
    sw      $s4, 0x20($sp)
    sw      $s3, 0x24($sp)
    sw      $s2, 0x28($sp)
    sw      $s1, 0x2C($sp)
    sw      $s0, 0x30($sp)
    sw      $t7, 0x34($sp)
    sw      $t6, 0x38($sp)
    sw      $t5, 0x3C($sp)
    sw      $t4, 0x40($sp)
    sw      $t3, 0x44($sp)
    sw      $t2, 0x48($sp)
    sw      $t1, 0x4C($sp)
    sw      $t0, 0x50($sp)
    sw      $a3, 0x54($sp)
    sw      $a2, 0x58($sp)
    sw      $a1, 0x5C($sp)
    sw      $a0, 0x60($sp)
    sw      $v1, 0x64($sp)
    sw      $v0, 0x68($sp)
    sw      $1, 0x6C($sp)
    mflo    $k0
    nop
    sw      $k0, 0x70($sp)
    mfhi    $k0
    nop
    sw      $k0, 0x74($sp)
    mfc0    $k0, C0_STATUS
    sll     $zero, 1
    sll     $zero, 1
    sll     $zero, 1
    sll     $zero, 1
    sw      $k0, 0x78($sp)
    mfc0    $k0, C0_EPC
    sll     $zero, 1
    sll     $zero, 1
    sll     $zero, 1
    sll     $zero, 1
    sw      $k0, 0x7C($sp)
    mfc0    $k0, C0_ERROREPC
    sll     $zero, 1
    sll     $zero, 1
    sll     $zero, 1
    sll     $zero, 1
    sw      $k0, 0x80($sp)
    
    li      $k1, M_CauseExcCode
    mfc0    $k0, C0_CAUSE
    and     $k0, $k1
    beq     $zero, $k0, _int
    nop
    j       _exception
    nop
    
_int:
    /* Initialise the kernel stack. */

    la      $k1, _exc_stack
    lw      $k0, 0($k1)
    sw      $sp, 0($k0)
    addiu   $k0, -4
    move    $sp, $k0

    /* Initialise the globals pointer. */

    lui $gp, %hi(_GLOBAL_OFFSET_TABLE_)
    ori $gp, $gp, %lo(_GLOBAL_OFFSET_TABLE_)

    /* Test with counter. */

    la      $t0, irq_counter
    lw      $t1, 0($t0)
    addiu   $t1, $t1, 1
    sw      $t1, 0($t0)

    /* Invoke the handler. */

    jal     irq_handle
    nop

    /* Restore the user stack. */

    addiu   $sp, 4
    lw      $sp, 0($sp)

    j       _exception_return
    nop
    
_exception:
    j       _exception_return
    nop

_exception_return:
    lw      $ra,    0($sp)
    lw      $fp,    4($sp)
    lw      $gp,    8($sp)
    lw      $t9,  0xC($sp)
    lw      $t8, 0x10($sp)
    lw      $s7, 0x14($sp)
    lw      $s6, 0x18($sp)
    lw      $s5, 0x1C($sp)
    lw      $s4, 0x20($sp)
    lw      $s3, 0x24($sp)
    lw      $s2, 0x28($sp)
    lw      $s1, 0x2C($sp)
    lw      $s0, 0x30($sp)
    lw      $t7, 0x34($sp)
    lw      $t6, 0x38($sp)
    lw      $t5, 0x3C($sp)
    lw      $t4, 0x40($sp)
    lw      $t3, 0x44($sp)
    lw      $t2, 0x48($sp)
    lw      $t1, 0x4C($sp)
    lw      $t0, 0x50($sp)
    lw      $a3, 0x54($sp)
    lw      $a2, 0x58($sp)
    lw      $a1, 0x5C($sp)
    lw      $a0, 0x60($sp)
    lw      $v1, 0x64($sp)
    lw      $v0, 0x68($sp)
    lw      $1, 0x6C($sp)
    lw      $k0, 0x70($sp)
    mtlo    $k0
    nop
    lw      $k0, 0x74($sp)
    mthi    $k0
    nop
    lw      $k0, 0x78($sp)
    mtc0    $k0, C0_STATUS
    nop
    sll     $zero, 1
    sll     $zero, 1
    sll     $zero, 1
    sll     $zero, 1
    lw      $k0, 0x7C($sp)
    mtc0    $k0, C0_EPC
    nop
    sll     $zero, 1
    sll     $zero, 1
    sll     $zero, 1
    sll     $zero, 1
    lw      $k0, 0x80($sp)
    mtc0    $k0, C0_ERROREPC
    nop
    sll     $zero, 1
    sll     $zero, 1
    sll     $zero, 1
    sll     $zero, 1
    addiu   $sp, 0x88       /* multiple of 8 */
    eret
    nop

.set reorder
.set at
